/* Mixmaster version 3  --  (C) 1999 Anonymizer Inc.

   Mixmaster may be redistributed and modified under certain conditions.
   This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF
   ANY KIND, either express or implied. See the file COPYRIGHT for
   details.

   DSA signatures without ASN.1 encoding
   $Id: dsasign.c,v 1.1 1999/03/20 21:07:23 um Exp $ */


/* SSLeay always encodes the signature values in a DER sequence */

#include <stdlib.h>
#include <stdio.h>
#include <bn.h>
#include "dsasign.h"

#ifndef DSA_F_DSA_DO_SIGN

DSA_SIG *DSA_SIG_new(void)
{
  DSA_SIG *ret;

  ret = malloc(sizeof(DSA_SIG));
  if (ret == NULL) return(NULL);
  ret->r = NULL;
  ret->s = NULL;
  return(ret);
}

void DSA_SIG_free(DSA_SIG *r)
{
  if (r == NULL) return;
  if (r->r) BN_clear_free(r->r);
  if (r->s) BN_clear_free(r->s);
  free(r);
}

DSA_SIG *DSA_do_sign(unsigned char *dgst, int dlen, DSA *dsa)
{
  BIGNUM *kinv=NULL,*r=NULL,*s=NULL,*m,*xr;
  BN_CTX *ctx=NULL;
  DSA_SIG *ret=NULL;

  m = BN_new();
  xr = BN_new();
  if ((s=BN_new()) == NULL) goto err;
  if ((ctx=BN_CTX_new()) == NULL) goto err;

  if ((dlen > BN_num_bytes(dsa->q)) || (dlen > 50)) goto err;
  
  if (!DSA_sign_setup(dsa,ctx,&kinv,&r)) goto err;
  if (BN_bin2bn(dgst,dlen,m) == NULL) goto err;

  /* s = k^-1 (m + xr) mod q */
  if (!BN_mod_mul(xr, dsa->priv_key, r, dsa->q, ctx)) goto err;
  if (!BN_add(m, m, xr)) goto err;
  if (!BN_mod_mul(s, kinv, m, dsa->q, ctx)) goto err;

  ret=DSA_SIG_new();
  ret->r = r;
  ret->s = s;

 err:
  BN_CTX_free(ctx);
  BN_clear_free(kinv);
  BN_clear_free(m);
  BN_clear_free(xr);
  return(ret);
}
#endif

