/* Mixmaster version 3  --  (C) 1999 Anonymizer Inc.

   Mixmaster may be redistributed and modified under certain conditions.
   This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF
   ANY KIND, either express or implied. See the file COPYRIGHT for
   details.

   Randomness
   $Id: random.c,v 1.32 1999/02/21 22:31:13 um Exp $ */


#include "mix3.h"
#include "crypto.h"
#ifdef POSIX
#include <sys/time.h>
#include <unistd.h>
#else
#include <io.h>
#include <process.h>
#endif
#include <assert.h>

BUFFER *rnd_buffer = NULL;

#ifdef USE_SSLEAY
int rnd_init(void)
{
  char r[PATHMAX];
  int n;
  LOCK *rndlock;

  rnd_buffer = buf_new();
  rndlock = lockfile(MIXRAND);
  mixfile(r, MIXRAND);
  n = RAND_load_file(r, 1024);
  if (n < 256 && rnd_seed() == -1)
    return (-1);
  rnd_update();
  RAND_write_file(r);
  unlockfile(rndlock);
  return (0);
}

int rnd_final(void)
{
  int err = 0;
  char r[PATHMAX];
  LOCK *rndlock;

  rnd_update();
  buf_free(rnd_buffer);
  rndlock = lockfile(MIXRAND);
  mixfile(r, MIXRAND);
  RAND_load_file(r, 1024);	/* add seed file again in case other instances
				   of the program have used it */
  if (RAND_write_file(r) < 1)
    err = -1;
  unlockfile(rndlock);
  RAND_cleanup();
  return (err);
}

int rnd_add(byte *b, long l)
{
  RAND_seed(b, l);
  return (0);
}
#endif

void rnd_time(void)
{
  int pid;

#ifdef WIN32
  SYSTEMTIME t;

#endif

#ifdef HAVE_GETTIMEOFDAY
  struct timeval tv;

  gettimeofday(&tv, 0);
  rnd_add((byte *) &tv, sizeof(tv));
#elif defined(WIN32)
  GetSystemTime(&t);
  rnd_add((byte *) &t, sizeof(t));
#else
  rnd_add((byte *) time(NULL), sizeof(time_t));
#endif
  pid = getpid();
  rnd_add((byte *) &pid, sizeof(pid));
}

void rnd_update(void)
{
  int fd = -1;
  byte b[512];

  rnd_time();
#ifdef DEV_URANDOM
  fd = open(DEV_URANDOM, O_RDONLY);
#endif
  if (fd != -1) {
    read(fd, b, sizeof(b));
    rnd_add(b, sizeof(b));
    close(fd);
  }
  rnd_add(rnd_buffer->data, rnd_buffer->length);
  buf_clear(rnd_buffer);
}

int rnd_bytes(byte *b, int n)
{
  RAND_bytes(b, n);
  return (0);
}

int rnd_number(int n)
{
  int r;

  assert(n > 0);
  if (n > 65535)
    do
      r = rnd_byte() * 65536 +
	rnd_byte() * 256 + rnd_byte();
    while (r >= n);
  else if (n > 255)
    do
      r = rnd_byte() * 256 + rnd_byte();
    while (r >= n);
  else
    do
      r = rnd_byte();
    while (r >= n);
  return r;
}

byte rnd_byte()
{
  byte b;

  rnd_bytes(&b, 1);
  return b;
}
