/* Mixmaster version 3  --  (C) 1999 Anonymizer Inc.

   Mixmaster may be redistributed and modified under certain conditions.
   This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF
   ANY KIND, either express or implied. See the file COPYRIGHT for
   details.

   Process remailer messages
   $Id: rem.c,v 1.38 1999/06/04 00:33:09 um Exp $ */


#include "mix3.h"
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef POSIX
#include <unistd.h>
#else
#include <io.h>
#endif
#ifndef _MSC
#include <dirent.h>
#endif
#include <assert.h>

int blockrequest(BUFFER *message);

#define REQUESTHELP 100
#define REQUESTSTATS 101
#define REQUESTKEY 200
#define REQUESTCONF 201
#define BLOCKREQUEST 666

#define CPUNKMSG 1
#define MIXMSG 2

int mix_decrypt(BUFFER *message)
{
  int type = 0;
  char *p, *q;
  BUFFER *field, *content;
  BUFFER *to, *subject, *replyto, *reply;
  FILE *f;
  BUFFER *block;
  int err = 0;

  mix_init(NULL);
  field = buf_new();
  content = buf_new();
  to = buf_new();
  replyto = buf_new();
  reply = buf_new();
  block = buf_new();
  subject = buf_new();
  buf_sets(subject, "Subject: Re: your mail");

  buf_rewind(message);

  f = mix_openfile(SOURCEBLOCK, "r");
  if (f != NULL) {
    buf_read(block, f);
    fclose(f);
  }
  for (;;) {
    err = buf_getheader(message, field, content);
    if (err == 1) {
      /* "::" marks for additional header lines */
      while (buf_lookahead(message, field) == 1)
	buf_getheader(message, field, content);
      if (isline(field, HDRMARK))
	continue;
      else
	goto hdrend;
    }
    if (err == -1)
      goto hdrend;
    p = field->data;
    q = content->data;

    if ((strieq(p, "from") || strieq(p, "sender")) &&
	doblock(content, block) != 0)
      goto end;

    if (strieq(p, "to"))
      buf_appends(to, q);
    else if (strieq(p, "from") && replyto->length == 0)
      /* reply to From address if no Reply-To header present */
      buf_sets(replyto, q);
    else if (strieq(p, "reply-to"))
      buf_sets(replyto, q);
    else if (MIX && strieq(p, "remailer-type") &&
	     strileft(q, "mixmaster"))
      type = MIXMSG;
    else if (strieq(p, "subject")) {
      if (strieq(q, "help") || strieq(q, "remailer-help"))
	type = REQUESTHELP;
      else if (strieq(q, "remailer-stats"))
	type = REQUESTSTATS;
      else if (strieq(q, "remailer-key"))
	type = REQUESTKEY;
      else if (strieq(q, "remailer-conf"))
	type = REQUESTCONF;
      else if (strileft(q, "destination-block"))
	type = BLOCKREQUEST;
      else {
	buf_sets(subject, "Subject: ");
	if (!strileft(q, "re:"))
	  buf_appends(subject, "Re: ");
	buf_appends(subject, q);
      }
    } else if (strieq(p, "test-to") || strieq(p, "encrypted") ||
	       strieq(p, "anon-to") || strieq(p, "request-remailing-to") ||
	       strieq(p, "remail-to") || strieq(p, "anon-post-to") ||
	       strieq(p, "post-to") || strieq(p, "anon-send-to") ||
	       strieq(p, "send-to") || strieq(p, "remix-to"))
      type = CPUNKMSG;
  }
hdrend:
  switch (type) {
  case REQUESTHELP:
    if (sendinfofile(HELPFILE, NULL, replyto, NULL) == -1)
      errlog(WARNING, "No help file available.\n");
    break;
  case REQUESTKEY:
    err = key(reply);
    if (err == 0)
      err = sendmail(reply, REMAILERNAME, replyto);
    break;
  case REQUESTSTATS:
    err = stats(reply);
    if (err == 0)
      err = sendmail(reply, REMAILERNAME, replyto);
    break;
  case REQUESTCONF:
    err = conf(reply);
    if (err == 0)
      err = sendmail(reply, REMAILERNAME, replyto);
    break;
  case CPUNKMSG:
    err = t1_decrypt(message);
    if (err != 0) {
      errlog(LOG, "Invalid type 1 message from %b\n", replyto);
      sendinfofile(USAGEFILE, USAGELOG, replyto, NULL);
      logmail(err == -2 ? MAILUSAGE : MAILERROR, message);
    }
    break;
  case MIXMSG:
    err = t2_decrypt(message);
    if (err == -1) {
      errlog(LOG, "Invalid type 2 message from %b\n", replyto);
      sendinfofile(USAGEFILE, USAGELOG, replyto, NULL);
      logmail(MAILERROR, message);
    }
    break;
  case BLOCKREQUEST:
    blockrequest(message);
    logmail(MAILBLOCK, message);
    break;
  default:
    if (strifind
	(replyto->data, "mailer-daemon")) {
      errlog(LOG, "Bounce mail from %b\n", replyto);
      logmail(MAILBOUNCE, message);
    } else if (strifind(to->data, REMAILERADDR) && blockrequest(message))
      logmail(MAILBLOCK, message);
    else if (!AUTOREPLY)
      logmail(MAILBOX, message);
    else if (strifind(to->data, REMAILERADDR)) {
      errlog(LOG, "Non-remailer message from %b\n", replyto);
      sendinfofile(USAGEFILE, USAGELOG, replyto, NULL);
      logmail(MAILUSAGE, message);
    } else if (strifind(to->data, COMPLAINTS)) {
      errlog(WARNING, "Abuse complaint from %b\n", replyto);
      sendinfofile(ABUSEFILE, NULL, replyto, subject);
      logmail(MAILABUSE, message);
    } else if (ANONADDR[0] && strifind(to->data, ANONADDR)) {
      errlog(LOG, "Reply to anonymous message from %b\n", replyto);
      sendinfofile(REPLYFILE, NULL, replyto, subject);
      logmail(MAILANON, message);
    } else {
      errlog(DEBUGINFO, "Mail from %b\n", replyto);
      logmail(MAILBOX, message);
    }
    err = 1;
  }
end:
  buf_free(field);
  buf_free(content);
  buf_free(to);
  buf_free(replyto);
  buf_free(reply);
  buf_free(block);
  buf_free(subject);
  return (err);
}

int t2_decrypt(BUFFER *in)
{
  int err = 0;
  BUFFER *msg;

  msg = buf_new();

  do {
    err = mix_dearmor(in, msg);
    if (err != -1) {
      if (v3_magic(msg->data))
	err = mix3_decrypt(msg);
      else
	err = mix2_decrypt(msg);
    }
  }
  while (in->ptr + 1000 < in->length);	/* accept several packets in one message */

  return (err);
}

int mix_pool(BUFFER *msg, int type, long latent)
{
  char path[PATHMAX], pathtmp[PATHMAX];
  FILE *f;
  int err = -1;

  f = pool_new(latent > 0 ? "lat" : "msg", pathtmp, path);
  if (f != NULL) {
    if (latent > 0)
      fprintf(f, "%d %ld\n", type, latent + time(NULL));
    else
      fprintf(f, "%d 0\n", type);
    err = buf_write(msg, f);
    fclose(f);
  }
  if (err == 0) {
    rename(pathtmp, path);
    errlog(DEBUGINFO, "Added message to pool.\n");
  }
  return (err);
}

int pool_packetfile(char *fname, BUFFER *mid, int packetnum)
     /* create a filename */
{
#ifdef SHORTNAMES
  sprintf(fname, "%s/p%02x%02x%02x%01x.%02x", POOLDIR,
	  mid->data[0], mid->data[1], mid->data[2], mid->data[3] & 15,
	  packetnum);
#else
  sprintf(fname, "%s/p%02x%02x%02x%02x%02x%02x%01x", POOLDIR,
	  packetnum, mid->data[0], mid->data[1], mid->data[2], mid->data[3],
	  mid->data[4], mid->data[5] & 15);
#endif
  return (0);
}

void pool_packetexp(void)
{
  DIR *d;
  struct dirent *e;
  struct stat sb;

  d = opendir(POOLDIR);
  if (d != NULL)
    for (;;) {
      e = readdir(d);
      if (e == NULL)
	break;
      if (e->d_name[0] == 'p') {
	if (stat(e->d_name, &sb) == 0 ||
	    time(NULL) - sb.st_mtime < PACKETEXP) {
	  errlog(NOTICE, "Expiring partial message %s.\n",
		 e->d_name);
	  unlink(e->d_name);
	}
      }
    }
  closedir(d);
}

void logmail(char *mailbox, BUFFER *message)
{
  /* mailbox is "|program", "user@host", "stdout" or "filename" */
  buf_rewind(message);
  if (mailbox[0] == '\0')	/* default action */
    mailbox = MAILBOX;
  if (strieq(mailbox, "stdout"))
    buf_write(message, stdout);
  else if (mailbox[0] == '|') {
    FILE *p;

    errlog(DEBUGINFO, "Piping message to %s.", mailbox + 1);
    p = openpipe(mailbox + 1);
    if (p != NULL) {
      buf_write(message, p);
      closepipe(p);
    }
  } else if (strchr(mailbox, '@')) {
    BUFFER *field, *content;

    field = buf_new();
    content = buf_new();
    while (buf_getheader(message, field, content) == 0)
      if (strieq(field->data, "x-loop") &&
	  strifind(content->data, REMAILERADDR)) {
	errlog(WARNING, "Loop detected! Message not sent to %s.\n", mailbox);
	goto isloop;
      }
    buf_sets(content, mailbox);
    sendmail(message, NULL, content);
  isloop:
    buf_free(field);
    buf_free(content);
  } else {
    FILE *mbox;

    mbox = mix_openfile(mailbox, "a");
    if (mbox == NULL) {
      errlog(ERRORMSG, "Can't write to mail folder %s\n", mailbox);
      return;
    }
    lock(mbox);
    if (!strileft(message->data, "From "))
      fprintf(mbox, "From Mixmaster Wed Jan 01 00:00:00 1970\n");
    buf_write(message, mbox);
    fprintf(mbox, "\n");
    unlock(mbox);
    fclose(mbox);
  }
}

int blockrequest(BUFFER *message)
{
  int request = 0, domain;
  BUFFER *from, *line, *field, *content, *addr;
  FILE *f;

  from = buf_new();
  line = buf_new();
  field = buf_new();
  content = buf_new();
  addr = buf_new();

  buf_rewind(message);
  while (buf_getheader(message, field, content) == 0)
    if (strieq(field->data, "from"))
      buf_set(from, content);
    else if (strieq(field->data, "subject"))
      buf_cat(message, content);
  while (buf_getline(message, line) != -1)
    if (strifind(line->data, "destination-block")) {
      buf_clear(addr);
      request = 1;
      if (strchr(line->data, '@')) {
	int c = 0;

	while (!strileft(line->data + line->ptr, "block"))
	  line->ptr++;
	while (c != ' ')
	  c = tolower(buf_getc(line));
	while (c == ' ')
	  c = buf_getc(line);
	do {
	  buf_appendc(addr, c);
	  c = buf_getc(line);
	} while (c > ' ');
      } else
	buf_set(addr, from);
      if (strieq(addr->data, REMAILERADDR)) {
	errlog(LOG, "Ignoring blocking request for %b from %b.\n", addr, from);
	return (2);
      }
      if (strieq(addr->data, from->data))
	errlog(NOTICE, "Blocking request for %b\n", addr);
      else
	errlog(NOTICE, "Blocking request for %b from %b\n", addr, from);
      if (AUTOBLOCK) {
	domain = 0;
	if (addr->data[0] == '@') {
	  domain = 1;
	  buf_sets(line, "postmaster");
	  buf_cat(line, addr);
	  buf_move(addr, line);
	}
	buf_clear(line);
	rfc822_addr(addr, line);
	if (line->length &&
	    strchr(line->data, '@') && strchr(strchr(line->data, '@'), '.')) {
	  f = mix_openfile(DESTBLOCK, "a");
	  if (f != NULL) {
	    lock(f);
	    sendinfofile(BLOCKFILE, NULL, line, NULL);
	    if (line->length) {
	      if (domain)
		fprintf(f, "%s", line->data + sizeof("postmaster") - 1);
	      else
		fprintf(f, "%s", line->data);
	    } else
	      errlog(NOTICE, "%b already blocked.\n", addr);
	    unlock(f);
	    fclose(f);
	  } else
	    errlog(ERRORMSG, "Can't write to %s.\n", DESTBLOCK);
	} else
	  errlog(WARNING, "Invalid address not added to %s: %b\n", DESTBLOCK,
		 addr);
      }
    }
  return (request);
}


int idexp(void)
{
  FILE *f;
  BUFFER *b;
  long now, then;
  char line[LINELEN];
  LOCK *i;

  if (IDEXP == 0)
    return (0);

  b = buf_new();
  f = mix_openfile(IDLOG, "r");
  if (f == NULL)
    return (-1);
  i = lockfile(IDLOG);
  now = time(NULL);
  fgets(line, sizeof(line), f);	/* replace first line */
  while (fgets(line, sizeof(line), f) != NULL) {
    sscanf(line, "%*s %ld", &then);
    if (now - then < IDEXP)
      buf_appends(b, line);
  }
  fclose(f);
  f = mix_openfile(IDLOG, "w");
  if (f != NULL) {
    fprintf(f, "expired %ld\n", now - IDEXP);
    buf_write(b, f);
    fclose(f);
  }
  unlockfile(i);
  return (0);
}
